/*
 * emdadm - manage Linux "emd" devices aka RAID arrays.
 *
 * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
 *
 * Added pluggable metadata support.
 * Copyright (C) 2004 Adaptec Inc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    Author: Neil Brown
 *    Email: <neilb@cse.unsw.edu.au>
 *    Paper: Neil Brown
 *           School of Computer Science and Engineering
 *           The University of New South Wales
 *           Sydney, 2052
 *           Australia
 */

#include	"mdadm.h"
#include	"md_u.h"

int Detail(char *dev, char *metaname, int brief, int test)
{	/*
	 * Print out details for an md array by using
	 * EMD_GET_ARRAY_INFO and EMD_GET_DISK_INFO ioctl calls
	 */
	mdu_array_info_t array;
	char *c;
        int rc;
	int err;
	
	int rv = test ? 4 : 1;

	memset(&array, 0, sizeof(mdu_array_info_t));
	if (strcmp(dev, "spares") == 0) {
		printf(" RaidDevice  Major  Minor    State\n");
		err = get_spare_info(&array, dev, metaname);
		if (err)
			printf("emadadm: Unable to get spare info\n");
		return err;
	}
        rc = get_array_info(dev, &array);
	if (rc != 0) {
                fprintf(stderr, Name ": cannot get array detail for %s: %s\n",
 				dev, strerror(errno));
		return rv;
	}

	rv = 0;
	c = map_num(pers, array.level);
	if (brief) 
		printf("ARRAY %s level=%s num-devices=%d", dev, c?c:"-unknown-",array.raid_disks );
	else 
		print_array_info(dev, &array);
        rv = print_disk_info(&array, brief, test, dev);

	/* If the requested device is a spare do not attempt to
	to get spare info again as it has been already obtained.
	XXX The check for RAID level should be changed to use a
	macro/enum instead of hardcoded values
	*/
	if (array.level != 0xfe) {
		err = get_spare_info(&array, dev, array.metaname);
		if (err)
			printf("emadadm: Unable to get spare info\n");
		return err;
	} else {
		return rv;
	}
}

int get_spare_info(mdu_array_info_t *array, char *dev, char *metaname)
{
	mdu_spare_disk_info_t spare_dev_list;
	int err, i;
	
	err = get_spares(dev, metaname, array->cookie, &spare_dev_list);
	if(err) {
		if (spare_dev_list.spare_info != NULL)
			free(spare_dev_list.spare_info);
		return err;
	}
	if (spare_dev_list.count == 0)
		return 0;
	for (i = 0; i < spare_dev_list.count; i++) {
		dump_disk_info(array, spare_dev_list.spare_info[i], 0, 0, dev);
		printf("\n");
	}
	if (spare_dev_list.spare_info != NULL)
		free(spare_dev_list.spare_info);
	return 0;
}

int get_spares(char *dev, char *metaname, unsigned long cookie,
	mdu_spare_disk_info_t *spare_dev_list)
{
	memset((void *)spare_dev_list, 0, sizeof(*spare_dev_list));
	spare_dev_list->parent_cookie = cookie;	
	memcpy(spare_dev_list->metaname, metaname, strlen(metaname));
	
	if(ioctl(mdctl_fd, EMD_GET_SPARES, spare_dev_list)){
		fprintf(stderr, Name ": EMD_GET_SPARES failed: %s\n",
			strerror(errno));
		return -EFAULT;
	}
	if (spare_dev_list->count == 0)
		return 0;
	spare_dev_list->spare_info = malloc(sizeof(mdu_disk_info_t)*
					spare_dev_list->count);
	memset(spare_dev_list->spare_info, 0, sizeof(mdu_disk_info_t)*
					spare_dev_list->count);
	if (spare_dev_list->spare_info == NULL)
		return -ENOMEM;
	
	if(ioctl(mdctl_fd, EMD_GET_SPARES, spare_dev_list)){
		fprintf(stderr, Name ": EMD_GET_SPARES failed: %s\n",
			strerror(errno));
		return -EFAULT;
	}
	
	return 0;
}


void print_array_info(char* dev, mdu_array_info_t *array)
{
	char *c;
	time_t atime;
        unsigned long events; 

	printf("%s:\n", dev);
	printf("        Version : %02d.%02d.%02d\n",
	       array->major_version, array->minor_version, array->patch_version);
	printf("  Metadata Type : %s\n", array->metaname);       
	atime = array->ctime;
	if (atime)
		printf("  Creation Time : %.24s\n", ctime(&atime));
	else	
		printf("  Creation Time : Unknown\n");
	c = map_num(pers, array->level);
	printf("     Raid Level : %s\n", c?c:"-unknown-");
 	printf("     Array Size :%s\n", 
		mb_size( (long long) array->size * (long long) 512) );
	printf("   Raid Devices : %d\n", array->raid_disks);
	printf("  Total Devices : %d\n", array->nr_disks);
	printf("Preferred Minor : %d\n", array->md_minor);
	atime = array->utime;
	if (atime)
		printf("    Update Time : %.24s\n", ctime(&atime));
	else
		printf("    Update Time : Unknown\n");
	if (array->size == 0)
		printf("          State : Array appears to be stopped.\n");
	else {
		switch (array->state & ~MDU_MSTATE_MODIFIER_MASK)
		{
		case MDU_MSTATE_UNKNOWN:
			printf( "          State : Unknown" );
			break;
		case MDU_MSTATE_OPTIMAL:
			printf( "          State : Optimal" );
			break;
		case MDU_MSTATE_DEGRADED:
			printf( "          State : Degraded" );
			break;
		case MDU_MSTATE_FAILED:
			printf( "          State : Failed" );
			break;
		case MDU_MSTATE_TEARDOWN:
			printf( "          State : Teardown" );
			break;
		}
			
		switch (array->state & MDU_MSTATE_MODIFIER_MASK)
		{
		case MDU_MSTATE_REBUILD_TARGET:
			printf( " - Rebuilding\n" );
			break;
		case MDU_MSTATE_INITIALIZING:
			printf( " - Initializing\n" );
			break;
		default:
			printf( "\n" );
			break;
		}
	}
			
	printf(" Active Devices : %d\n", array->active_disks);
	printf("Working Devices : %d\n", array->working_disks);
	printf(" Failed Devices : %d\n", array->failed_disks);
	printf("  Spare Devices : %d\n", array->spare_disks);
	events = get_gencount( dev );
	printf("         Events : %ld\n", events);
	
	if (array->level == 5) {
		c = map_num(r5layout, array->layout);
		printf("         Layout : %s\n", c?c:"-unknown-");
	}
	switch (array->level) {
	case 0:
	case 4:
	case 5:
		printf("     Chunk Size : %dK\n", array->chunk_size/1024);
		break;
	case -1:
		printf("       Rounding : %dK\n", array->chunk_size/1024);
		break;
	default: break;
	}
	
	printf("\n");
	
	{
		struct mdstat_ent *ms = mdstat_read();
		struct mdstat_ent *e;
		
		for (e = ms; e; e = e->next)
			if (e->devnum == array->md_minor >> EMD_MINOR_SHIFT) {
				if (e->percent >= 0)
					printf(" %s status : %d%% complete\n\n", e->syncitem, e->percent);
				break;
			}
		free_mdstat(ms);
	}
return;
}


int print_disk_info(mdu_array_info_t *array, int brief, int test, char *array_name)
{
	int d;
	char *devices = NULL;
	int spares = 0;
        int rc;
	unsigned long parent_cookie;
	mdu_disk_info_t disk;
	int rv = test ? 4 : 1;
	
        parent_cookie = array->cookie;
	if (!brief)
		printf(" RaidDevice  Major  Minor    State\n");

	for (d = 0; d < (signed int)(array->raid_disks + array->spare_disks); d++) {
                rc = get_disk_info_indexed(parent_cookie, d, &disk);
		if ( rc != 0) {
			if (d < array->raid_disks) 
				disk.state = MDU_MSTATE_FAILED;
			else
				continue;
		}
		if (d >= array->raid_disks &&
		    disk.major == 0 &&
		    disk.minor == 0)
			continue;
		dump_disk_info(array, disk, 0, 0, array_name);
		if (!brief) printf("\n");
	}
	if (spares && brief) printf(" spares=%d", spares);
	if (brief && devices) printf("\n   devices=%s", devices);
	if (brief) printf("\n");
	if (test && (rv&2)) rv &= ~1;
	return rv;
}
int dump_disk_info(mdu_array_info_t *array, mdu_disk_info_t disk, 
		   int brief, int test, char *array_name)
{
	char *devices = NULL;
	int fd;
	int d;
	unsigned long disk_size;
	unsigned long long ldisk_size;
	int rv = test ? 4 : 1;
	char *dv;

	if (!brief) {
		printf("%5d     %5d   %5d     ", 
		       disk.number, disk.major, disk.minor);
			       
		switch (disk.state & ~MDU_MSTATE_MODIFIER_MASK)
			{
			case MDU_MSTATE_OPTIMAL:
				printf( " Optimal" );
				break;
			case MDU_MSTATE_DEGRADED:
				printf( " Degraded" );
				break;
			case MDU_MSTATE_FAILED:
				printf( " Failed" );
				break;
			case MDU_MSTATE_TEARDOWN:
				printf( " Teardown" );
				break;
			case 0:
				printf( " Spare  " );
				break;
			}
		
			switch (disk.state & MDU_MSTATE_MODIFIER_MASK)
			{
			case MDU_MSTATE_REBUILD_TARGET:
				if (array->state 
				     != MDU_MSTATE_DEGRADED)
					printf(" Initializing");
				else
					printf( " Rebuilding ");
				break;
			default:
				printf("            ");
				break;
			}
	}
	
	if (test && d < array->raid_disks && disk.state & (1<<MD_DISK_FAULTY)) {
		if ((rv & 1) && (array->level ==4 || array->level == 5))
			rv |= 2;
		rv |= 1;
	}
	if ((dv = map_dev(disk.major, disk.minor))) {
		if (brief) {
			if (devices) {
				devices = realloc(devices,
					  strlen(devices)+1+strlen(dv)+1);
				strcat(strcat(devices,","),dv);
			} else
				devices = strdup(dv);
		} else {
			printf("  %s", dv);
		}
		
		fd = open(dv, O_RDWR);
#ifdef BLKGETSIZE64
		if (ioctl(fd, BLKGETSIZE64, &ldisk_size)==0)
			;
		else
#endif
		if (ioctl(fd, BLKGETSIZE, &disk_size)==0) {
			ldisk_size = disk_size;
			ldisk_size <<= 9;
		}
		else 
			ldisk_size = 0;
		if (!brief) {
			if (ldisk_size)
				printf(" %s", mb_size(ldisk_size));
			if (strcmp(dv, array_name) == 0 )
				printf( " <== this device" );
		}
		close(fd);

	}
	return 0;
}
