/*
 * emdadm - manage Linux "emd" devices aka RAID arrays.
 *
 * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
 *
 * Added pluggable metadata support, monitoring of initializing arrays and
 * commented support for spares.
 * Copyright (C) 2004 Adaptec Inc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    Author: Neil Brown
 *    Email: <neilb@cse.unsw.edu.au>
 *    Paper: Neil Brown
 *           School of Computer Science and Engineering
 *           The University of New South Wales
 *           Sydney, 2052
 *           Australia
 */

#include	"mdadm.h"
#include	"md_p.h"
#include	"md_u.h"
#include	"dlink.h"
#include	<sys/wait.h>
#include	<sys/signal.h>
#include	<values.h>


static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd);

static char *percentalerts[] = { 
	"Started",
	"20%",
	"40%",
	"60%",
	"80%",
};

int Monitor(mddev_dev_t devlist,
	    char *mailaddr, char *alert_cmd,
	    int period, int daemonise, int scan, int oneshot,
	    char *config, int test)
{
	/*
	 * Every few seconds, scan every md device looking for changes
	 * When a change is found, log it, possibly run the alert command,
	 * and possibly send Email
	 *
	 * For each array, we record:
	 *   Update time
	 *   active/working/failed/spare drives
	 *   State of each device.
	 *   %rebuilt if rebuilding
	 *
	 * If the update time changes, check out all the data again
	 * It is possible that we cannot get the state of each device
	 * due to bugs in the md kernel module.
	 * We also read /proc/emdstat to get rebuild percent,
	 * and to get state on all active devices incase of kernel bug.
	 *
	 * Events are:
	 *    Fail
	 *	An active device had Faulty set or Active/Sync removed
	 *    FailSpare
	 *      A spare device had Faulty set
	 *    SpareActive
	 *      An active device had a reverse transition
	 *    RebuildStarted
	 *      percent went from -1 to +ve
	 *    Rebuild20 Rebuild40 Rebuild60 Rebuild80
	 *      percent went from below to not-below that number
	 *    DeviceDisappeared
	 *      Couldn't access a device which was previously visible
	 *
	 * if we detect an array with active<raid and spare==0
	 * we look at other arrays that have same spare-group
	 * If we find one with active==raid and spare>0,
	 *  and if we can get_disk_info and find a name
	 *  Then we hot-remove and hot-add to the other array
	 *
	 * If devlist is NULL, then we can monitor everything because --scan
	 * was given.  We get an initial list from config file and add anything
	 * that appears in /proc/emdstat
	 */

	struct state {
		char *devname;
		int devnum;	/* to sync with mdstat info */
		long utime;
		int err;
		char *spare_group;
		int active, working, failed, spare, raid;
		int expected_spares;
		char pattern[MD_SB_DISKS];
		int devid[MD_SB_DISKS];
		int percent;
		char syncitem[80];
		struct state *next;
	} *statelist = NULL;
	int finished = 0;
	struct mdstat_ent *mdstat = NULL;
	char tmpstring[160];				
	mdu_array_info_t array;

	if (!mailaddr) {
		mailaddr = conf_get_mailaddr(config);
		if (mailaddr)
			fprintf(stderr, Name ": Monitor using email address \"%s\" from config file\n",
			       mailaddr);
	}
	if (!alert_cmd) {
		alert_cmd = conf_get_program(config);
		if (alert_cmd)
			fprintf(stderr, Name ": Monitor using program \"%s\" from config file\n",
			       alert_cmd);
	}

	if (daemonise) {
		int pid = fork();
		if (pid > 0) {
			printf("%d\n", pid);
			return 0;
		}
		if (pid < 0) {
			perror("daemonise");
			return 1;
		}
		close(0);
		open("/dev/null", 3);
		dup2(0,1);
		dup2(0,2);
		setsid();
	}

	if (devlist == NULL) {
		mddev_ident_t mdlist = conf_get_ident(config, NULL);
		if (mdlist)
			fprintf(stderr, "Monitoring devices/arrays from "
				"config file\n");
			
		for (; mdlist; mdlist=mdlist->next) {
                        if (get_array_info(mdlist->devname, &array ) != 0) {
				fprintf(stderr, "%s not present\n",
					mdlist->devname);
				continue;
			}
			struct state *st = malloc(sizeof *st);
			if (st == NULL)
				continue;
			st->devname = strdup(mdlist->devname);
			st->utime = 0;
			st->next = statelist;
			st->err = 0;
			st->devnum = get_unitnum(st->devname);
			st->percent = -2;
			st->expected_spares = mdlist->spare_disks;
			if (mdlist->spare_group)
				st->spare_group = strdup(mdlist->spare_group);
			else
				st->spare_group = NULL;
			statelist = st;
		}
	} else {
		mddev_dev_t dv;
		for (dv=devlist ; dv; dv=dv->next) {
                        if (get_array_info(dv->devname, &array ) != 0) {
				fprintf(stderr, "%s not present\n",
					dv->devname);
				continue;
			}
			struct state *st = malloc(sizeof *st);
			if (st == NULL)
				continue;
			st->devname = strdup(dv->devname);
			st->utime = 0;
			st->next = statelist;
			st->err = 0;
			st->devnum = get_unitnum(st->devname);
			st->percent = -2;
			st->expected_spares = -1;
			st->spare_group = NULL;
			statelist = st;
		}
	}
	if (!scan && statelist == NULL) {
		fprintf(stderr, "No valid devices in command line or "
			"config file. Try --scan option\n");
		return -1;
	} else if (scan)
		fprintf(stderr, "Monitoring all arrays in /proc/emdstat and "
			"new arrays that will be added later\n");


	while (! finished) {
		int new_found = 0;
		struct state *st;
		int i;

		if (mdstat)
			free_mdstat(mdstat);
		mdstat = mdstat_read();

		for (st=statelist; st; st=st->next) {
			mdu_array_info_t array;
			struct mdstat_ent *mse;
			char *dev = st->devname;

			if (test)
				alert("TestMessage", dev, NULL, mailaddr, alert_cmd);
                        if ( get_array_info( dev, &array ) != 0) {
				if (!st->err)
					alert("DeviceDisappeared", dev, NULL,
					      mailaddr, alert_cmd);
/*					fprintf(stderr, Name ": cannot get array info for %s: %s\n",
						dev, strerror(errno));
*/				st->err=1;
				continue;
			}
			if (array.level != 1 && array.level != 5 && array.level != -4) {
				if (!st->err)
					alert("DeviceDisappeared", dev, "Wrong-Level",
					      mailaddr, alert_cmd);
				st->err = 1;
				continue;
			}

			for (mse = mdstat ; mse ; mse=mse->next)
				if (mse->devnum == st->devnum){
					mse->devnum = MAXINT; /* flag it as "used" */
					break;
				}


			if (st->utime == array.utime &&
			    st->failed == array.failed_disks &&
			    st->working == array.working_disks &&
			    st->spare == array.spare_disks &&
			    (mse == NULL  || (
				    mse->percent == st->percent
				    ))) {
				st->err = 0;
				continue;
			}
			if (st->utime == 0 && /* new array */
			    mse &&	/* is in /proc/emdstat */
			    mse->pattern && strchr(mse->pattern, '_') /* degraded */
			     && strcmp(st->pattern, mse->pattern)
				)
				alert("DegradedArray", dev, NULL, mailaddr, alert_cmd);

			if (mse &&
			    st->percent == -1 && 
			    mse->percent >= 0) {
				sprintf(tmpstring, "%s started", mse->syncitem);
				alert(tmpstring, dev, NULL, mailaddr, 
				      alert_cmd);
			}
			if (mse &&
			    st->percent >= 0 &&
			    mse->percent >= 0 &&
			    (mse->percent / 20) > (st->percent / 20)) {

				sprintf(tmpstring, "%s %s complete",
					mse->syncitem,
					percentalerts[mse->percent/20]);
				alert(tmpstring, dev, NULL, mailaddr, 
				      alert_cmd);
			}

			if (mse &&
			    mse->percent == -1 &&
			    st->percent >= 0) {
				sprintf(tmpstring, "%s finished",
					st->syncitem);
				alert(tmpstring, dev, NULL, mailaddr, 
				      alert_cmd);
			}

			if (mse) {
				st->percent = mse->percent;
				strcpy(st->syncitem, mse->syncitem);
			}
			for (i=0; i < array.raid_disks; i++) {
				int change;
				char *dv = NULL;

				change = mse->pattern[i] ^ st->pattern[i];
				if (change && !st->err) {
					if (mse->pattern[i] == '_') 
						alert("Fail", dev, dv,
						       mailaddr, alert_cmd);
				}	
			}
			strcpy(st->pattern, mse->pattern);
			st->active = array.active_disks;
			st->working = array.working_disks;
			st->spare = array.spare_disks;
			st->failed = array.failed_disks;
			st->utime = array.utime;
			st->raid = array.raid_disks;
			st->err = 0;
		}
		/* now check if there are any new devices found in mdstat */
		if (scan) {
			struct mdstat_ent *mse;
			for (mse=mdstat; mse; mse=mse->next) 
				if (mse->devnum != MAXINT &&
				    (strcmp(mse->level, "raid1")==0 ||
				     strcmp(mse->level, "raid5")==0 ||
				     strcmp(mse->level, "multipath")==0)
					) {
					struct state *st = malloc(sizeof *st);
					mdu_array_info_t array;
					if (st == NULL)
						continue;
					st->devname = strdup(get_md_name(mse->devnum));
					if (get_array_info( st->devname, &array ) != 0 ) {
						/* no such array */
						free(st->devname);
						free(st);
						continue;
					}
					st->utime = 0;
					st->next = statelist;
					st->err = 1;
					st->devnum = mse->devnum;
					st->percent = -2;
					st->spare_group = NULL;
					st->expected_spares = -1;
					statelist = st;
					alert("NewArray", st->devname, NULL, mailaddr, alert_cmd);
					new_found = 1;
				}
		}
		if (!new_found) {
			if (oneshot)
				break;
			else
				sleep(period);
		}
		test = 0;
	}
	return 0;
}


static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd)
{
#if 0
	fprintf(stderr, "%s event had been detected on md device %s.\n\n", event, dev);
#endif
	if (!cmd && !mailaddr) {
		time_t now = time(0);
	       
		printf("%1.15s: %s on %s %s\n", ctime(&now)+4, event, dev, disc?disc:"");
	}
	if (cmd) {
		int pid = fork();
		switch(pid) {
		default:
			waitpid(pid, NULL, 0);
			break;
		case -1:
			break;
		case 0:
			execl(cmd, cmd, event, dev, disc, NULL);
			exit(2);
		}
	}
	if (mailaddr && 
	    (strncmp(event, "Fail", 4)==0 || 
	     strncmp(event, "Test", 4)==0 ||
	     strncmp(event, "Degrade", 7)==0)) {
		FILE *mp = popen(Sendmail, "w");
		if (mp) {
			char hname[256];
			gethostname(hname, sizeof(hname));
			signal(SIGPIPE, SIG_IGN);
			fprintf(mp, "From: " Name " monitoring <root>\n");
			fprintf(mp, "To: %s\n", mailaddr);
			fprintf(mp, "Subject: %s event on %s:%s\n\n", event, dev, hname);

			fprintf(mp, "This is an automatically generated mail message from " Name "\n");
			fprintf(mp, "running on %s\n\n", hname);

			fprintf(mp, "%s event had been detected on md device %s.\n\n", event, dev);

			if (disc)
				fprintf(mp, "It could be related to component device %s.\n\n", disc);

			fprintf(mp, "Faithfully yours, etc.\n");
			fclose(mp);
		}

	}
	/* FIXME log the event to syslog maybe */
}
